@@ -20,22 +20,45 @@
 
 import sys
 import os
+import gi
+from gi.repository import Gio
 from gi.repository import GLib
+gi.require_version('Gtk', '3.0')
 from gi.repository import Gtk
+gi.require_version('IBus', '1.0')
 from gi.repository import IBus
 import locale
 import gettext
 import config
-import subprocess
 from keycapturedialog import KeyCaptureDialog
 
 _ = lambda a : gettext.dgettext(config.gettext_package, a)
 
+
+def get_hangul_keyboard_list():
+    from ctypes import CDLL, c_int, c_char_p
+    libhangul = CDLL('libhangul.so.1')
+    libhangul.hangul_ic_get_n_keyboards.argtypes = []
+    libhangul.hangul_ic_get_n_keyboards.restype = c_int
+    libhangul.hangul_ic_get_keyboard_id.argtypes = [c_int]
+    libhangul.hangul_ic_get_keyboard_id.restype = c_char_p
+    libhangul.hangul_ic_get_keyboard_name.argtypes = [c_int]
+    libhangul.hangul_ic_get_keyboard_name.restype = c_char_p
+
+    n = libhangul.hangul_ic_get_n_keyboards()
+    list = []
+    for i in range(n):
+        id = libhangul.hangul_ic_get_keyboard_id(i).decode('UTF-8')
+        name = libhangul.hangul_ic_get_keyboard_name(i).decode('UTF-8')
+        list.append((id, name))
+    return list
+
+
 class Setup ():
     def __init__ (self, bus):
         self.__bus = bus
-        self.__config = self.__bus.get_config()
-        self.__config.connect("value-changed", self.on_value_changed, None)
+        self.__settings = Gio.Settings(schema="org.freedesktop.ibus.engine.hangul")
+        self.__settings.connect("changed", self.on_value_changed)
 
         ui_file = os.path.join(os.path.dirname(__file__), "setup.ui")
         self.__builder = Gtk.Builder()
@@ -43,41 +66,60 @@ class Setup ():
         self.__builder.add_from_file(ui_file)
 
         # Hangul tab
-        pipe = subprocess.Popen([config.setuplibdir + '/hangul_keyboard_list'], stdout = subprocess.PIPE)
-        list = pipe.communicate()[0].split('\n')
+        list = get_hangul_keyboard_list()
         
         self.__hangul_keyboard = self.__builder.get_object("HangulKeyboard")
         model = Gtk.ListStore(str, str, int)
         i = 0
-        for line in list:
-            items = line.split('\t')
-            if len(items) > 1:
-                model.append([items[1], items[0], i])
-                i+=1
+        for (id, name) in list:
+            model.append([name, id, i])
+            i+=1
 
         self.__hangul_keyboard.set_model(model)
         renderer = Gtk.CellRendererText()
         self.__hangul_keyboard.pack_start(renderer, True)
         self.__hangul_keyboard.add_attribute(renderer, "text", 0)
 
-        default = GLib.Variant.new_string("2")
-        current = self.__read("HangulKeyboard", default).get_string()
+        current = self.__read("hangul-keyboard").get_string()
         for i in model:
             if i[1] == current:
                 self.__hangul_keyboard.set_active(i[2])
                 break
 
-        self.__word_commit = self.__builder.get_object("WordCommit")
+        self.__start_in_hangul_mode = self.__builder.get_object("StartInHangulMode")
+        initial_input_mode = self.__read("initial-input-mode").get_string()
+        self.__start_in_hangul_mode.set_active(initial_input_mode == "hangul")
 
-        default = GLib.Variant.new_boolean(False)
-        word_commit = self.__read("WordCommit", default).get_boolean()
+        self.__word_commit = self.__builder.get_object("WordCommit")
+        word_commit = self.__read("word-commit").get_boolean()
         self.__word_commit.set_active(word_commit)
 
         self.__auto_reorder = self.__builder.get_object("AutoReorder")
-        default = GLib.Variant.new_boolean(True)
-        auto_reorder = self.__read("AutoReorder", default).get_boolean()
+        auto_reorder = self.__read("auto-reorder").get_boolean()
         self.__auto_reorder.set_active(auto_reorder)
 
+        button = self.__builder.get_object("HangulKeyListAddButton")
+        button.connect("clicked", self.on_hangul_key_add, None)
+
+        button = self.__builder.get_object("HangulKeyListRemoveButton")
+        button.connect("clicked", self.on_hangul_key_remove, None)
+
+        model = Gtk.ListStore(str)
+
+        keylist_str = self.__read("switch-keys").get_string()
+        self.__hangul_key_list_str = keylist_str.split(',')
+        for i in self.__hangul_key_list_str:
+            model.append([i])
+
+        self.__hangul_key_list = self.__builder.get_object("HangulKeyList")
+        self.__hangul_key_list.set_model(model)
+        column = Gtk.TreeViewColumn()
+        column.set_title("key")
+        renderer = Gtk.CellRendererText()
+        column.pack_start(renderer, True)
+        column.add_attribute(renderer, "text", 0)
+        self.__hangul_key_list.append_column(column)
+
         # hanja tab
         button = self.__builder.get_object("HanjaKeyListAddButton")
         button.connect("clicked", self.on_hanja_key_add, None)
@@ -87,8 +129,7 @@ class Setup ():
 
         model = Gtk.ListStore(str)
 
-        default = GLib.Variant.new_string("Hangul_Hanja,F9")
-        keylist_str = self.__read("HanjaKeys", default).get_string()
+        keylist_str = self.__read("hanja-keys").get_string()
         self.__hanja_key_list_str = keylist_str.split(',')
         for i in self.__hanja_key_list_str:
             model.append([i])
@@ -110,28 +151,49 @@ class Setup ():
         self.__window = self.__builder.get_object("SetupDialog")
         icon_file = os.path.join(config.pkgdatadir, "icons", "ibus-hangul.svg")
         self.__window.set_icon_from_file(icon_file)
-        self.__window.connect("response", self.on_response, None)
         self.__window.show()
 
-        ok_button = self.__builder.get_object("button_cancel")
-        ok_button.grab_focus()
+        button = self.__builder.get_object("button_apply")
+        button.connect("clicked", self.on_apply, None)
+
+        button = self.__builder.get_object("button_cancel")
+        button.connect("clicked", self.on_cancel, None)
+        #button.grap_focus()
+
+        button = self.__builder.get_object("button_ok")
+        button.connect("clicked", self.on_ok, None)
 
     def run(self):
-        res = self.__window.run()
-        if (res == Gtk.ResponseType.OK):
-            self.on_ok()
-        self.__window.destroy()
+        Gtk.main()
 
     def apply(self):
         model = self.__hangul_keyboard.get_model()
         i = self.__hangul_keyboard.get_active()
-        self.__write("HangulKeyboard", GLib.Variant.new_string(model[i][1]))
+        self.__write("hangul-keyboard", GLib.Variant.new_string(model[i][1]))
+
+        start_in_hangul_mode = self.__start_in_hangul_mode.get_active()
+        if start_in_hangul_mode:
+            self.__write("initial-input-mode", GLib.Variant.new_string("hangul"))
+        else:
+            self.__write("initial-input-mode", GLib.Variant.new_string("latin"))
 
         word_commit = self.__word_commit.get_active()
-        self.__write("WordCommit", GLib.Variant.new_boolean(word_commit))
+        self.__write("word-commit", GLib.Variant.new_boolean(word_commit))
 
         auto_reorder = self.__auto_reorder.get_active()
-        self.__write("AutoReorder", GLib.Variant.new_boolean(auto_reorder))
+        self.__write("auto-reorder", GLib.Variant.new_boolean(auto_reorder))
+
+        model = self.__hangul_key_list.get_model()
+        str = ""
+        iter = model.get_iter_first()
+        while iter:
+            if len(str) > 0:
+                str += ","
+                str += model.get_value(iter, 0)
+            else:
+                str += model.get_value(iter, 0)
+            iter = model.iter_next(iter)
+        self.__write("switch-keys", GLib.Variant.new_string(str))
 
         model = self.__hanja_key_list.get_model()
         str = ""
@@ -143,20 +205,52 @@ class Setup ():
             else:
                 str += model.get_value(iter, 0)
             iter = model.iter_next(iter)
-        self.__write("HanjaKeys", GLib.Variant.new_string(str))
+        self.__write("hanja-keys", GLib.Variant.new_string(str))
 
-    def on_response(self, widget, id, data = None):
-        if id == Gtk.ResponseType.APPLY:
-            self.apply()
-            widget.emit_stop_by_name("response")
-        if id == Gtk.ResponseType.NONE:
-            widget.emit_stop_by_name("response")
-
-    def on_ok(self):
+    def on_apply(self, widget, data):
         self.apply()
 
+    def on_cancel(self, widget, data):
+        self.__window.destroy()
+        Gtk.main_quit()
+
+    def on_ok(self, widget, data):
+        self.apply()
+        self.__window.destroy()
+        Gtk.main_quit()
+
+    def on_hangul_key_add(self, widget, data = None):
+        dialog = KeyCaptureDialog(_("Select Hangul toggle key"), self.__window)
+        dialog.set_markup(_("Press any key which you want to use as hangul toggle key. "
+                "The key you pressed is displayed below.\n"
+                "If you want to use it, click \"Ok\" or click \"Cancel\""))
+        res = dialog.run()
+        if res == Gtk.ResponseType.OK:
+            key_str = dialog.get_key_string()
+            if len(key_str) > 0:
+                model = self.__hangul_key_list.get_model()
+                iter = model.get_iter_first()
+                while iter:
+                    str = model.get_value(iter, 0)
+                    if str == key_str:
+                        model.remove(iter)
+                        break
+                    iter = model.iter_next(iter)
+
+                model.append([key_str])
+        dialog.destroy()
+
+    def on_hangul_key_remove(self, widget, data = None):
+        selection = self.__hangul_key_list.get_selection()
+        (model, iter) = selection.get_selected()
+        if model and iter:
+            model.remove(iter)
+
     def on_hanja_key_add(self, widget, data = None):
         dialog = KeyCaptureDialog(_("Select Hanja key"), self.__window)
+        dialog.set_markup(_("Press any key which you want to use as hanja key. "
+                "The key you pressed is displayed below.\n"
+                "If you want to use it, click \"Ok\" or click \"Cancel\""))
         res = dialog.run()
         if res == Gtk.ResponseType.OK:
             key_str = dialog.get_key_string()
@@ -179,29 +273,31 @@ class Setup ():
         if model and iter:
             model.remove(iter)
 
-    def on_value_changed(self, config, section, name, value, data):
-        if section == "engine/Hangul":
-            if name == "HangulKeyboard":
-                model = self.__hangul_keyboard.get_model()
-                for i in model:
-                    if i[1] == value:
-                        self.__hangul_keyboard.set_active(i[2])
-                        break
-            elif name == "HanjaKeys":
-                self.__hanja_key_list_str = value.split(',')
+    def on_value_changed(self, settings, key):
+        value = settings.get_value(key)
+        if key == "hangul-keyboard":
+            model = self.__hangul_keyboard.get_model()
+            for i in model:
+                if i[1] == value.get_string():
+                    self.__hangul_keyboard.set_active(i[2])
+                    break
+        elif key == "switch-keys":
+            self.__hangul_key_list_str = value.get_string().split(',')
+        elif key == "hanja-keys":
+            self.__hanja_key_list_str = value.get_string().split(',')
 
-    def __read(self, name, v):
-        value = self.__config.get_value("engine/Hangul", name)
-        if value is None:
-            return v
-        return value
+    def __read(self, key):
+        return self.__settings.get_value(key)
 
-    def __write(self, name, v):
-        return self.__config.set_value("engine/Hangul", name, v)
+    def __write(self, key, v):
+        self.__settings.set_value(key, v)
 
 if __name__ == "__main__":
     locale.bindtextdomain(config.gettext_package, config.localedir)
-    locale.bind_textdomain_codeset(config.gettext_package, "UTF-8")
+    gettext.bind_textdomain_codeset(config.gettext_package, "UTF-8")
+
+    GLib.set_prgname("ibus-setup-hangul")
+    GLib.set_application_name(_("IBusHangul Setup"))
 
     bus = IBus.Bus()
     if bus.is_connected():
