@@ -23,6 +23,7 @@
 #endif
 
 #include <ibus.h>
+#include <gio/gio.h>
 #include <hangul.h>
 #include <string.h>
 #include <ctype.h>
@@ -38,18 +39,19 @@ typedef struct _IBusHangulEngineClass IBusHangulEngineClass;
 typedef struct _HotkeyList HotkeyList;
 
 enum {
-    INPUT_MODE_DIRECT,
     INPUT_MODE_HANGUL,
+    INPUT_MODE_LATIN,
     INPUT_MODE_COUNT,
 };
 
 struct _IBusHangulEngine {
-    IBusEngine parent;
+    IBusEngineSimple parent;
 
     /* members */
     HangulInputContext *context;
     UString* preedit;
     int input_mode;
+    unsigned int input_purpose;
     gboolean hanja_mode;
     HanjaList* hanja_list;
     int last_lookup_method;
@@ -64,7 +66,7 @@ struct _IBusHangulEngine {
 };
 
 struct _IBusHangulEngineClass {
-    IBusEngineClass parent;
+    IBusEngineSimpleClass parent;
 };
 
 struct KeyEvent {
@@ -136,6 +138,10 @@ static void ibus_hangul_engine_candidate_clicked
                                              guint                   index,
                                              guint                   button,
                                              guint                   state);
+static void ibus_hangul_engine_set_content_type
+                                            (IBusEngine             *engine,
+                                             guint                   purpose,
+                                             guint                   hints);
 
 static void ibus_hangul_engine_flush        (IBusHangulEngine       *hangul);
 static void ibus_hangul_engine_clear_preedit_text
@@ -147,7 +153,7 @@ static void ibus_hangul_engine_update_lookup_table
                                             (IBusHangulEngine       *hangul);
 static gboolean ibus_hangul_engine_has_preedit
                                             (IBusHangulEngine       *hangul);
-static void ibus_hangul_engine_toggle_input_mode
+static void ibus_hangul_engine_switch_input_mode
                                             (IBusHangulEngine       *hangul);
 static void ibus_hangul_engine_set_input_mode
                                             (IBusHangulEngine       *hangul,
@@ -163,10 +169,8 @@ static bool ibus_hangul_engine_on_transition
                                              const ucschar          *preedit,
                                              void                   *data);
 
-static void ibus_config_value_changed       (IBusConfig             *config,
-                                             const gchar            *section,
-                                             const gchar            *name,
-                                             GVariant               *value,
+static void        settings_changed         (GSettings              *settings,
+                                             const gchar            *key,
                                              gpointer                user_data);
 
 static void        lookup_table_set_visible (IBusLookupTable        *table,
@@ -193,10 +197,11 @@ static gboolean hotkey_list_has_modifier    (HotkeyList           *list,
 
 static glong ucschar_strlen (const ucschar* str);
 
-static IBusEngineClass *parent_class = NULL;
+static IBusEngineSimpleClass *parent_class = NULL;
 static HanjaTable *hanja_table = NULL;
 static HanjaTable *symbol_table = NULL;
-static IBusConfig *config = NULL;
+static GSettings *settings_hangul = NULL;
+static GSettings *settings_panel = NULL;
 static GString    *hangul_keyboard = NULL;
 static HotkeyList hanja_keys;
 static HotkeyList switch_keys;
@@ -206,7 +211,12 @@ static int lookup_table_orientation = 0;
 static IBusKeymap *keymap = NULL;
 static gboolean word_commit = FALSE;
 static gboolean auto_reorder = TRUE;
-static int initial_input_mode = INPUT_MODE_DIRECT;
+static gboolean disable_latin_mode = FALSE;
+static int initial_input_mode = INPUT_MODE_LATIN;
+/**
+ * whether to use event forwarding workaround
+ */
+static gboolean use_event_forwarding = TRUE;
 
 static glong
 ucschar_strlen (const ucschar* str)
@@ -235,7 +245,7 @@ ibus_hangul_engine_get_type (void)
     };
 
     if (type == 0) {
-            type = g_type_register_static (IBUS_TYPE_ENGINE,
+            type = g_type_register_static (IBUS_TYPE_ENGINE_SIMPLE,
                                            "IBusHangulEngine",
                                            &type_info,
                                            (GTypeFlags) 0);
@@ -247,97 +257,102 @@ ibus_hangul_engine_get_type (void)
 void
 ibus_hangul_init (IBusBus *bus)
 {
-    GVariant* value;
+    GVariant* value = NULL;
 
     hanja_table = hanja_table_load (NULL);
 
     symbol_table = hanja_table_load (IBUSHANGUL_DATADIR "/data/symbol.txt");
 
-    config = ibus_bus_get_config (bus);
-    if (config)
-        g_object_ref_sink (config);
+    settings_hangul = g_settings_new ("org.freedesktop.ibus.engine.hangul");
+    settings_panel = g_settings_new ("org.freedesktop.ibus.panel");
 
-    hangul_keyboard = g_string_new_len ("2", 8);
-    value = ibus_config_get_value (config, "engine/hangul",
-                                         "hangul-keyboard");
+    hangul_keyboard = g_string_new_len (NULL, 8);
+    value = g_settings_get_value (settings_hangul, "hangul-keyboard");
     if (value != NULL) {
         const gchar* str = g_variant_get_string (value, NULL);
         g_string_assign (hangul_keyboard, str);
-        g_variant_unref(value);
+        g_clear_pointer (&value, g_variant_unref);
     }
 
     hotkey_list_init(&switch_keys);
 
-    value = ibus_config_get_value (config, "engine/hangul",
-                                         "switch-keys");
+    value = g_settings_get_value (settings_hangul, "switch-keys");
     if (value != NULL) {
         const gchar* str = g_variant_get_string (value, NULL);
         hotkey_list_set_from_string(&switch_keys, str);
-        g_variant_unref(value);
+        g_clear_pointer (&value, g_variant_unref);
     } else {
-	hotkey_list_append(&switch_keys, IBUS_Hangul, 0);
-	hotkey_list_append(&switch_keys, IBUS_space, IBUS_SHIFT_MASK);
+        hotkey_list_append(&switch_keys, IBUS_Hangul, 0);
+        hotkey_list_append(&switch_keys, IBUS_space, IBUS_SHIFT_MASK);
     }
 
     hotkey_list_init(&hanja_keys);
 
-    value = ibus_config_get_value (config, "engine/hangul",
-                                         "hanja-keys");
+    value = g_settings_get_value (settings_hangul, "hanja-keys");
     if (value != NULL) {
         const gchar* str = g_variant_get_string (value, NULL);
         hotkey_list_set_from_string(&hanja_keys, str);
-        g_variant_unref(value);
+        g_clear_pointer (&value, g_variant_unref);
     } else {
-	hotkey_list_append(&hanja_keys, IBUS_Hangul_Hanja, 0);
-	hotkey_list_append(&hanja_keys, IBUS_F9, 0);
+        hotkey_list_append(&hanja_keys, IBUS_Hangul_Hanja, 0);
+        hotkey_list_append(&hanja_keys, IBUS_F9, 0);
     }
 
     hotkey_list_init (&on_keys);
-    value = ibus_config_get_value (config, "engine/hangul", "on-keys");
+    value = g_settings_get_value (settings_hangul, "on-keys");
     if (value != NULL) {
         const gchar* str = g_variant_get_string (value, NULL);
         hotkey_list_set_from_string (&on_keys, str);
-        g_variant_unref (value);
-    } else {
+        g_clear_pointer (&value, g_variant_unref);
     }
 
     hotkey_list_init (&off_keys);
-    value = ibus_config_get_value (config, "engine/hangul", "off-keys");
+    value = g_settings_get_value (settings_hangul, "off-keys");
     if (value != NULL) {
         const gchar* str = g_variant_get_string (value, NULL);
         hotkey_list_set_from_string (&off_keys, str);
-        g_variant_unref (value);
-    } else {
-        hotkey_list_append (&off_keys, IBUS_KEY_Escape, 0);
+        g_clear_pointer (&value, g_variant_unref);
     }
 
-    value = ibus_config_get_value (config, "engine/hangul",
-                                         "word-commit");
+    value = g_settings_get_value (settings_hangul, "word-commit");
     if (value != NULL) {
         word_commit = g_variant_get_boolean (value);
-        g_variant_unref(value);
+        g_clear_pointer (&value, g_variant_unref);
     }
 
-    value = ibus_config_get_value (config, "engine/hangul", "auto-reorder");
+    value = g_settings_get_value (settings_hangul, "auto-reorder");
     if (value != NULL) {
         auto_reorder = g_variant_get_boolean (value);
-        g_variant_unref (value);
+        g_clear_pointer (&value, g_variant_unref);
     }
 
-    value = ibus_config_get_value (config, "engine/hangul", "initial-input-mode");
+    value = g_settings_get_value (settings_hangul, "disable-latin-mode");
+    if (value != NULL) {
+        disable_latin_mode = g_variant_get_boolean (value);
+        g_clear_pointer (&value, g_variant_unref);
+    }
+
+    value = g_settings_get_value (settings_hangul, "initial-input-mode");
     if (value != NULL) {
         const gchar* str = g_variant_get_string (value, NULL);
         if (strcmp(str, "latin") == 0) {
-            initial_input_mode = INPUT_MODE_DIRECT;
+            initial_input_mode = INPUT_MODE_LATIN;
         } else if (strcmp(str, "hangul") == 0) {
             initial_input_mode = INPUT_MODE_HANGUL;
         }
-        g_variant_unref (value);
+        g_clear_pointer (&value, g_variant_unref);
     }
 
-    value = ibus_config_get_value (config, "panel", "lookup-table-orientation");
+    value = g_settings_get_value (settings_hangul, "use-event-forwarding");
+    if (value != NULL) {
+        use_event_forwarding = g_variant_get_boolean (value);
+        g_clear_pointer (&value, g_variant_unref);
+    }
+
+    value = g_settings_get_value (settings_panel, "lookup-table-orientation");
     if (value != NULL) {
         lookup_table_orientation = g_variant_get_int32(value);
+        g_clear_pointer (&value, g_variant_unref);
     }
 
     keymap = ibus_keymap_get("us");
@@ -362,8 +377,8 @@ ibus_hangul_exit (void)
     hanja_table_delete (symbol_table);
     symbol_table = NULL;
 
-    g_object_unref (config);
-    config = NULL;
+    g_clear_object (&settings_hangul);
+    g_clear_object (&settings_panel);
 
     g_string_free (hangul_keyboard, TRUE);
     hangul_keyboard = NULL;
@@ -376,7 +391,7 @@ ibus_hangul_engine_class_init (IBusHangulEngineClass *klass)
     IBusObjectClass *ibus_object_class = IBUS_OBJECT_CLASS (klass);
     IBusEngineClass *engine_class = IBUS_ENGINE_CLASS (klass);
 
-    parent_class = (IBusEngineClass *) g_type_class_peek_parent (klass);
+    parent_class = (IBusEngineSimpleClass *) g_type_class_peek_parent (klass);
 
     object_class->constructor = ibus_hangul_engine_constructor;
     ibus_object_class->destroy = (IBusObjectDestroyFunc) ibus_hangul_engine_destroy;
@@ -399,6 +414,7 @@ ibus_hangul_engine_class_init (IBusHangulEngineClass *klass)
     engine_class->property_activate = ibus_hangul_engine_property_activate;
 
     engine_class->candidate_clicked = ibus_hangul_engine_candidate_clicked;
+    engine_class->set_content_type = ibus_hangul_engine_set_content_type;
 }
 
 static void
@@ -416,9 +432,14 @@ ibus_hangul_engine_init (IBusHangulEngine *hangul)
     hangul->preedit = ustring_new();
     hangul->hanja_list = NULL;
     hangul->input_mode = initial_input_mode;
+    hangul->input_purpose = IBUS_INPUT_PURPOSE_FREE_FORM;
     hangul->hanja_mode = FALSE;
     hangul->last_lookup_method = LOOKUP_METHOD_PREFIX;
 
+    if (disable_latin_mode) {
+        hangul->input_mode = INPUT_MODE_HANGUL;
+    }
+
     hangul->prop_list = ibus_prop_list_new ();
     g_object_ref_sink (hangul->prop_list);
 
@@ -431,7 +452,7 @@ ibus_hangul_engine_init (IBusHangulEngine *hangul)
                               tooltip,
                               TRUE, TRUE, PROP_STATE_UNCHECKED, NULL);
     symbol = ibus_hangul_engine_get_input_mode_symbol (hangul,
-                                                       initial_input_mode);
+                                                       hangul->input_mode);
     ibus_property_set_symbol(prop, symbol);
     g_object_ref_sink (prop);
     ibus_prop_list_append (hangul->prop_list, prop);
@@ -462,8 +483,10 @@ ibus_hangul_engine_init (IBusHangulEngine *hangul)
     hangul->table = ibus_lookup_table_new (9, 0, TRUE, FALSE);
     g_object_ref_sink (hangul->table);
 
-    g_signal_connect (config, "value-changed",
-                      G_CALLBACK(ibus_config_value_changed), hangul);
+    g_signal_connect (settings_hangul, "changed",
+                      G_CALLBACK (settings_changed), hangul);
+    g_signal_connect (settings_panel, "changed",
+                      G_CALLBACK (settings_changed), hangul);
 }
 
 static GObject*
@@ -708,7 +731,7 @@ h_ibus_text_get_substring (IBusText* ibus_text, glong p1, glong p2)
 static HanjaList*
 ibus_hangul_engine_lookup_hanja_table (const char* key, int method)
 {
-    HanjaList* list;
+    HanjaList* list = NULL;
 
     if (key == NULL)
         return NULL;
@@ -1035,6 +1058,10 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
     if (keyval == IBUS_Shift_L || keyval == IBUS_Shift_R)
         return FALSE;
 
+    // On password mode, we ignore hotkeys
+    if (hangul->input_purpose == IBUS_INPUT_PURPOSE_PASSWORD)
+        return IBUS_ENGINE_CLASS (parent_class)->process_key_event (engine, keyval, keycode, modifiers);
+
     // If a hotkey has any modifiers, we ignore that modifier
     // keyval, or we cannot make the hanja key work.
     // Because when we get the modifier key alone, we commit the
@@ -1046,7 +1073,7 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
         return FALSE;
 
     if (hotkey_list_match(&switch_keys, keyval, modifiers)) {
-        ibus_hangul_engine_toggle_input_mode (hangul);
+        ibus_hangul_engine_switch_input_mode (hangul);
         return TRUE;
     }
 
@@ -1055,13 +1082,13 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
         return FALSE;
     }
 
-    if (hangul->input_mode == INPUT_MODE_DIRECT)
-        return FALSE;
+    if (hangul->input_mode == INPUT_MODE_LATIN)
+        return IBUS_ENGINE_CLASS (parent_class)->process_key_event (engine, keyval, keycode, modifiers);
 
     /* This feature is for vi* users.
      * On Esc, the input mode is changed to latin */
     if (hotkey_list_match (&off_keys, keyval, modifiers)) {
-        ibus_hangul_engine_set_input_mode (hangul, INPUT_MODE_DIRECT);
+        ibus_hangul_engine_set_input_mode (hangul, INPUT_MODE_LATIN);
         /* If we return TRUE, then vi will not receive "ESC" key event. */
         return FALSE;
     }
@@ -1190,6 +1217,57 @@ ibus_hangul_engine_process_key_event (IBusEngine     *engine,
             ibus_hangul_engine_flush (hangul);
     }
 
+    /* We always return TRUE here even if we didn't use this event.
+     * Instead, we forward the event to clients.
+     *
+     * Because IBus has a problem with sync mode.
+     * I think it's limitations of IBus implementation.
+     * We call several engine functions(updating preedit text and committing
+     * text) inside this function.
+     * But clients cannot receive the results of other calls until this
+     * function ends. Clients only get one result from a remote call at a time
+     * because clients may run on event loop.
+     * Clients may process this event first and then get the results which
+     * may change the preedit text or commit text.
+     * So the event order is broken.
+     * Call order:
+     *      engine                          client
+     *                                      call process_key_event
+     *      begin process_key_event
+     *        call commit_text
+     *        call update_preedit_text
+     *      return the event as unused
+     *                                      receive the result of process_key_event
+     *                                      receive the result of commit_text
+     *                                      receive the result of update_preedit_text
+     *
+     * To solve this problem, we return TRUE as if we consumed this event.
+     * After that, we forward this event to clients.
+     * Then clients may get the events in correct order.
+     * This approach is a kind of async processing.
+     * Call order:
+     *      engine                          client
+     *                                      call process_key_event
+     *      begin process_key_event
+     *        call commit_text
+     *        call update_preedit_text
+     *        call forward_key_event
+     *      return the event as used
+     *                                      receive the result of process_key_event
+     *                                      receive the result of commit_text
+     *                                      receive the result of update_preedit_text
+     *                                      receive the forwarded key event
+     *
+     * See: https://github.com/choehwanjin/ibus-hangul/issues/40
+     */
+    if (use_event_forwarding) {
+        if (!retval) {
+            ibus_engine_forward_key_event (engine, keyval, keycode, modifiers);
+        }
+
+        return TRUE;
+    }
+
     return retval;
 }
 
@@ -1212,6 +1290,7 @@ ibus_hangul_engine_flush (IBusHangulEngine *hangul)
 	str = ustring_begin (hangul->preedit);
 	text = ibus_text_new_from_ucs4 (str);
 
+        g_debug("flush: %s", text->text);
 	ibus_engine_commit_text ((IBusEngine *) hangul, text);
 
 	ustring_clear(hangul->preedit);
@@ -1245,7 +1324,7 @@ ibus_hangul_engine_focus_in (IBusEngine *engine)
         ibus_hangul_engine_update_lookup_table_ui (hangul);
     }
 
-    parent_class->focus_in (engine);
+    IBUS_ENGINE_CLASS (parent_class)->focus_in (engine);
 }
 
 static void
@@ -1264,7 +1343,7 @@ ibus_hangul_engine_focus_out (IBusEngine *engine)
         ibus_engine_hide_auxiliary_text (engine);
     }
 
-    parent_class->focus_out ((IBusEngine *) hangul);
+    IBUS_ENGINE_CLASS (parent_class)->focus_out ((IBusEngine *) hangul);
 }
 
 static void
@@ -1273,13 +1352,13 @@ ibus_hangul_engine_reset (IBusEngine *engine)
     IBusHangulEngine *hangul = (IBusHangulEngine *) engine;
 
     ibus_hangul_engine_flush (hangul);
-    parent_class->reset (engine);
+    IBUS_ENGINE_CLASS (parent_class)->reset (engine);
 }
 
 static void
 ibus_hangul_engine_enable (IBusEngine *engine)
 {
-    parent_class->enable (engine);
+    IBUS_ENGINE_CLASS (parent_class)->enable (engine);
 
     ibus_engine_get_surrounding_text (engine, NULL, NULL, NULL);
 }
@@ -1288,19 +1367,19 @@ static void
 ibus_hangul_engine_disable (IBusEngine *engine)
 {
     ibus_hangul_engine_focus_out (engine);
-    parent_class->disable (engine);
+    IBUS_ENGINE_CLASS (parent_class)->disable (engine);
 }
 
 static void
 ibus_hangul_engine_page_up (IBusEngine *engine)
 {
-    parent_class->page_up (engine);
+    IBUS_ENGINE_CLASS (parent_class)->page_up (engine);
 }
 
 static void
 ibus_hangul_engine_page_down (IBusEngine *engine)
 {
-    parent_class->page_down (engine);
+    IBUS_ENGINE_CLASS (parent_class)->page_down (engine);
 }
 
 static void
@@ -1313,7 +1392,7 @@ ibus_hangul_engine_cursor_up (IBusEngine *engine)
         ibus_hangul_engine_update_lookup_table_ui (hangul);
     }
 
-    parent_class->cursor_up (engine);
+    IBUS_ENGINE_CLASS (parent_class)->cursor_up (engine);
 }
 
 static void
@@ -1326,7 +1405,7 @@ ibus_hangul_engine_cursor_down (IBusEngine *engine)
         ibus_hangul_engine_update_lookup_table_ui (hangul);
     }
 
-    parent_class->cursor_down (engine);
+    IBUS_ENGINE_CLASS (parent_class)->cursor_down (engine);
 }
 
 static void
@@ -1344,7 +1423,7 @@ ibus_hangul_engine_property_activate (IBusEngine    *engine,
     } else if (strcmp(prop_name, "InputMode") == 0) {
         IBusHangulEngine *hangul = (IBusHangulEngine *) engine;
 
-        ibus_hangul_engine_toggle_input_mode (hangul);
+        ibus_hangul_engine_switch_input_mode (hangul);
     } else if (strcmp(prop_name, "hanja_mode") == 0) {
         IBusHangulEngine *hangul = (IBusHangulEngine *) engine;
 
@@ -1380,14 +1459,13 @@ ibus_hangul_engine_has_preedit (IBusHangulEngine *hangul)
 }
 
 static void
-ibus_hangul_engine_toggle_input_mode (IBusHangulEngine *hangul)
+ibus_hangul_engine_switch_input_mode (IBusHangulEngine *hangul)
 {
-    int input_mode;
+    int input_mode = hangul->input_mode + 1;
 
-    if (hangul->input_mode == INPUT_MODE_HANGUL)
-        input_mode = INPUT_MODE_DIRECT;
-    else
+    if (input_mode >= INPUT_MODE_COUNT) {
         input_mode = INPUT_MODE_HANGUL;
+    }
 
     ibus_hangul_engine_set_input_mode (hangul, input_mode);
 }
@@ -1399,10 +1477,10 @@ ibus_hangul_engine_get_input_mode_symbol (IBusHangulEngine *hangul,
     IBusText **symbols = hangul->input_mode_symbols;
 
     if (symbols[0] == NULL) {
-        symbols[INPUT_MODE_DIRECT] = ibus_text_new_from_string ("A");
-        g_object_ref_sink(symbols[INPUT_MODE_DIRECT]);
         symbols[INPUT_MODE_HANGUL] = ibus_text_new_from_string ("한");
         g_object_ref_sink(symbols[INPUT_MODE_HANGUL]);
+        symbols[INPUT_MODE_LATIN] = ibus_text_new_from_string ("EN");
+        g_object_ref_sink(symbols[INPUT_MODE_LATIN]);
     }
 
     if (input_mode >= INPUT_MODE_COUNT)
@@ -1419,9 +1497,14 @@ ibus_hangul_engine_set_input_mode (IBusHangulEngine *hangul, int input_mode)
 
     ibus_hangul_engine_flush (hangul);
 
+    if (disable_latin_mode) {
+        return;
+    }
+
     prop = hangul->prop_hangul_mode;
 
     hangul->input_mode = input_mode;
+    g_debug("input_mode: %s", (input_mode == INPUT_MODE_HANGUL) ? "hangul" : "latin");
 
     symbol = ibus_hangul_engine_get_input_mode_symbol (hangul, input_mode);
     ibus_property_set_symbol(prop, symbol);
@@ -1457,41 +1540,57 @@ ibus_hangul_engine_on_transition (HangulInputContext     *hic,
 }
 
 static void
-ibus_config_value_changed (IBusConfig   *config,
-                           const gchar  *section,
-                           const gchar  *name,
-                           GVariant     *value,
-                           gpointer      user_data)
+settings_changed (GSettings    *settings,
+                  const gchar  *key,
+                  gpointer      user_data)
 {
     IBusHangulEngine *hangul = (IBusHangulEngine *) user_data;
+    GValue schema_value = G_VALUE_INIT;
+    const gchar *schema_id;
+    GVariant *value;
 
-    if (strcmp(section, "engine/hangul") == 0) {
-        if (strcmp(name, "hangul-keyboard") == 0) {
+    g_return_if_fail (G_IS_SETTINGS (settings));
+
+    g_value_init (&schema_value, G_TYPE_STRING);
+    g_object_get_property (G_OBJECT (settings), "schema-id", &schema_value);
+    schema_id = g_value_get_string (&schema_value);
+    value = g_settings_get_value (settings, key);
+    if (strcmp (schema_id, "org.freedesktop.ibus.engine.hangul") == 0) {
+        if (strcmp(key, "hangul-keyboard") == 0) {
             const gchar *str = g_variant_get_string(value, NULL);
             g_string_assign (hangul_keyboard, str);
             hangul_ic_select_keyboard (hangul->context, hangul_keyboard->str);
-        } else if (strcmp(name, "hanja-keys") == 0) {
+        } else if (strcmp (key, "hanja-keys") == 0) {
             const gchar* str = g_variant_get_string(value, NULL);
 	    hotkey_list_set_from_string(&hanja_keys, str);
-        } else if (strcmp(name, "word-commit") == 0) {
+        } else if (strcmp (key, "word-commit") == 0) {
             word_commit = g_variant_get_boolean (value);
-        } else if (strcmp (name, "auto-reorder") == 0) {
+        } else if (strcmp (key, "auto-reorder") == 0) {
             auto_reorder = g_variant_get_boolean (value);
-        } else if (strcmp (name, "switch-keys") == 0) {
+        } else if (strcmp (key, "switch-keys") == 0) {
             const gchar* str = g_variant_get_string(value, NULL);
 	    hotkey_list_set_from_string(&switch_keys, str);
-        } else if (strcmp (name, "on-keys") == 0) {
+        } else if (strcmp (key, "on-keys") == 0) {
             const gchar* str = g_variant_get_string(value, NULL);
 	    hotkey_list_set_from_string(&on_keys, str);
-        } else if (strcmp (name, "off-keys") == 0) {
+        } else if (strcmp (key, "off-keys") == 0) {
             const gchar* str = g_variant_get_string(value, NULL);
 	    hotkey_list_set_from_string(&off_keys, str);
+        } else if (strcmp (key, "initial-input-mode") == 0) {
+            const gchar* str = g_variant_get_string (value, NULL);
+            if (strcmp(str, "latin") == 0) {
+                initial_input_mode = INPUT_MODE_LATIN;
+            } else if (strcmp(str, "hangul") == 0) {
+                initial_input_mode = INPUT_MODE_HANGUL;
+            }
         }
-    } else if (strcmp(section, "panel") == 0) {
-        if (strcmp(name, "lookup-table-orientation") == 0) {
+    } else if (strcmp (schema_id, "org.freedesktop.ibus.panel") == 0) {
+        if (strcmp (key, "lookup-table-orientation") == 0) {
             lookup_table_orientation = g_variant_get_int32(value);
         }
     }
+    g_variant_unref (value);
+    g_value_unset (&schema_value);
 }
 
 static void
@@ -1562,6 +1661,18 @@ ibus_hangul_engine_candidate_clicked (IBusEngine     *engine,
     }
 }
 
+static void
+ibus_hangul_engine_set_content_type (IBusEngine     *engine,
+                                     guint           purpose,
+                                     guint           hints)
+{
+    IBusHangulEngine *hangul = (IBusHangulEngine *) engine;
+    if (hangul == NULL)
+        return;
+
+    hangul->input_purpose = purpose;
+}
+
 static void
 hotkey_list_init(HotkeyList* list)
 {
