@@ -59,10 +59,25 @@ static void
 start_component (void)
 {
     IBusComponent *component;
+    IBusConfig* config;
+    gboolean res;
 
     ibus_init ();
 
     bus = ibus_bus_new ();
+
+    res = ibus_bus_is_connected (bus);
+    if (!res) {
+        g_warning ("Unable to connect to IBus");
+        exit (2);
+    }
+
+    config = ibus_bus_get_config (bus);
+    if (config == NULL) {
+        g_warning ("Unable to connect to IBus config component");
+        exit (3);
+    }
+
     g_signal_connect (bus, "disconnected", G_CALLBACK (ibus_disconnected_cb), NULL);
 
     ibus_hangul_init (bus);
@@ -117,13 +132,21 @@ main (gint argc, gchar **argv)
 
     context = g_option_context_new ("- ibus hangul engine component");
 
-    g_option_context_add_main_entries (context, entries, "ibus-hangul");
+    g_option_context_add_main_entries (context, entries, GETTEXT_PACKAGE);
 
     if (!g_option_context_parse (context, &argc, &argv, &error)) {
         g_print ("Option parsing failed: %s\n", error->message);
         exit (-1);
     }
 
+    if (verbose) {
+        const gchar* value = g_getenv ("G_MESSAGES_DEBUG");
+        if (value == NULL) {
+            g_setenv ("G_MESSAGES_DEBUG", "all", TRUE);
+        }
+    }
+
     start_component ();
+
     return 0;
 }
